@@ -77,6 +77,42 @@ typedef struct {
 	gboolean enable_background;
 } LsmSvgViewBackground;
 
+cairo_operator_t cairo_operators[] = {
+	CAIRO_OPERATOR_CLEAR,
+	CAIRO_OPERATOR_SOURCE,
+	CAIRO_OPERATOR_DEST,
+	CAIRO_OPERATOR_OVER,
+	CAIRO_OPERATOR_DEST_OVER,
+	CAIRO_OPERATOR_IN,
+	CAIRO_OPERATOR_DEST_IN,
+	CAIRO_OPERATOR_OUT,
+	CAIRO_OPERATOR_DEST_OUT,
+	CAIRO_OPERATOR_ATOP,
+	CAIRO_OPERATOR_DEST_ATOP,
+	CAIRO_OPERATOR_XOR,
+	CAIRO_OPERATOR_ADD,
+	CAIRO_OPERATOR_MULTIPLY,
+	CAIRO_OPERATOR_SCREEN,
+	CAIRO_OPERATOR_OVERLAY,
+	CAIRO_OPERATOR_DARKEN,
+	CAIRO_OPERATOR_LIGHTEN,
+	CAIRO_OPERATOR_COLOR_DODGE,
+	CAIRO_OPERATOR_COLOR_BURN,
+	CAIRO_OPERATOR_HARD_LIGHT,
+	CAIRO_OPERATOR_SOFT_LIGHT,
+	CAIRO_OPERATOR_DIFFERENCE,
+	CAIRO_OPERATOR_EXCLUSION
+};
+
+void
+lsm_cairo_set_comp_op (cairo_t *cairo, LsmSvgCompOp comp_op)
+{
+	if (G_LIKELY (cairo != NULL && comp_op >= LSM_SVG_COMP_OP_CLEAR && comp_op <= LSM_SVG_COMP_OP_EXCLUSION)) {
+		lsm_log_render ("[lsmSvgView::set_comp_op] Set comp-op to %s", lsm_svg_comp_op_to_string (comp_op));
+		cairo_set_operator (cairo, cairo_operators[comp_op]);
+	}
+}
+
 double
 lsm_svg_view_normalize_length (LsmSvgView *view, const LsmSvgLength *length, LsmSvgLengthDirection direction)
 {
@@ -85,6 +121,27 @@ lsm_svg_view_normalize_length (LsmSvgView *view, const LsmSvgLength *length, Lsm
 	return lsm_svg_length_normalize (length, view->viewbox_stack->data, view->style->font_size_px, direction);
 }
 
+double *
+lsm_svg_view_normalize_length_list (LsmSvgView *view, const LsmSvgLengthList *list, LsmSvgLengthDirection direction, unsigned int *n_data)
+{
+	double *data;
+	unsigned int i;
+
+	g_return_val_if_fail (n_data != NULL, NULL);
+	*n_data = 0;
+	g_return_val_if_fail (LSM_IS_SVG_VIEW (view), NULL);
+
+	if (list->n_lengths == 0)
+		return NULL;
+
+	*n_data = list->n_lengths;
+	data = g_new (double, list->n_lengths);
+	for (i = 0; i < list->n_lengths; i++)
+		data[i] = lsm_svg_view_normalize_length (view, &list->lengths[i], direction);
+
+	return data;
+}
+
 static void
 _start_pattern (LsmSvgView *view, const LsmBox *extents, const LsmBox *object_extents, double opacity)
 {
@@ -660,7 +717,8 @@ paint (LsmSvgView *view, LsmSvgViewPathInfos *path_infos)
 	cairo = view->dom_view.cairo;
 	style = view->style;
 
-	if (style->opacity != NULL &&
+	if ((style->opacity != NULL ||
+	     style->comp_op->value != LSM_SVG_COMP_OP_SRC_OVER ) &&
 	    style->ignore_group_opacity &&
 	    g_strcmp0 (style->filter->value, "none") == 0) {
 		group_opacity = style->opacity->value;
@@ -671,15 +729,17 @@ paint (LsmSvgView *view, LsmSvgViewPathInfos *path_infos)
 			     style->stroke->paint.type == LSM_SVG_PAINT_TYPE_URI_RGB_COLOR ||
 			     style->fill->paint.type == LSM_SVG_PAINT_TYPE_URI_RGB_COLOR ||
 			     style->fill->paint.type == LSM_SVG_PAINT_TYPE_URI) &&
-			group_opacity < 1.0;
+			(group_opacity < 1.0 || style->comp_op->value != LSM_SVG_COMP_OP_SRC_OVER );
 	} else {
 		use_group = FALSE;
 		group_opacity = 1.0;
 	}
 
 	/* Instead of push_group, we should restrict to the current path bounding box */
-	if (use_group)
+	if (use_group) {
 		cairo_push_group (cairo);
+	} else if (style->comp_op->value != LSM_SVG_COMP_OP_SRC_OVER)
+		lsm_cairo_set_comp_op (cairo, style->comp_op->value);
 
 	if (_set_color (view,
 			path_infos,
@@ -762,8 +822,13 @@ paint (LsmSvgView *view, LsmSvgViewPathInfos *path_infos)
 
 	if (use_group) {
 		cairo_pop_group_to_source (cairo);
+		if (G_UNLIKELY (style->comp_op->value != LSM_SVG_COMP_OP_SRC_OVER))
+			lsm_cairo_set_comp_op (cairo, style->comp_op->value);
 		cairo_paint_with_alpha (cairo, group_opacity);
 	}
+
+	if (view->style->comp_op->value != LSM_SVG_COMP_OP_SRC_OVER)
+		lsm_cairo_set_comp_op (cairo, LSM_SVG_COMP_OP_SRC_OVER);
 }
 
 static void
@@ -779,6 +844,38 @@ process_path (LsmSvgView *view, LsmSvgViewPathInfos *path_infos)
 		paint (view, path_infos);
 }
 
+void
+lsm_svg_view_show_viewport (LsmSvgView*view, const LsmBox *viewport)
+{
+	LsmSvgPaint *paint;
+
+	g_return_if_fail (LSM_IS_SVG_VIEW (view));
+	g_return_if_fail (viewport != NULL);
+
+	paint = &view->style->viewport_fill->paint;
+
+	switch (paint->type) {
+		case LSM_SVG_PAINT_TYPE_RGB_COLOR:
+			cairo_set_source_rgba (view->dom_view.cairo,
+					       paint->color.red,
+					       paint->color.green,
+					       paint->color.blue,
+					       view->style->viewport_fill_opacity->value);
+			break;
+		case LSM_SVG_PAINT_TYPE_CURRENT_COLOR:
+			cairo_set_source_rgba (view->dom_view.cairo,
+					       view->style->color->value.red,
+					       view->style->color->value.green,
+					       view->style->color->value.blue,
+					       view->style->viewport_fill_opacity->value);
+		default:
+			return;
+	}
+
+	cairo_rectangle (view->dom_view.cairo, viewport->x, viewport->y, 0, 0);
+	cairo_paint (view->dom_view.cairo);
+}
+
 /*
  * Code for show_rectangle and show ellipse is inspired from
  * the librsvg library (rsvg-shapes.c)
@@ -961,7 +1058,10 @@ lsm_svg_view_show_polygon (LsmSvgView *view, const char *points)
 }
 
 static void
-_update_pango_layout (LsmSvgView *view, char const *string, double x, double y, LsmSvgViewPathInfos *path_infos)
+_update_pango_layout (LsmSvgView *view, unsigned int n, char const *string, double x, double y,
+		      unsigned int n_dx, const double *dx,
+		      unsigned int n_dy, const double *dy,
+		      LsmSvgViewPathInfos *path_infos)
 {
 	const LsmSvgStyle *style;
 	PangoLayout *pango_layout;
@@ -970,7 +1070,10 @@ _update_pango_layout (LsmSvgView *view, char const *string, double x, double y,
 	PangoStyle font_style;
 	PangoLayoutIter *iter;
 	PangoRectangle rectangle;
+	PangoAttrList *attrs;
+	PangoAttribute *attr;
 	int baseline;
+	int i;
 	double x1, y1;
 
 	style = view->style;
@@ -1028,7 +1131,36 @@ _update_pango_layout (LsmSvgView *view, char const *string, double x, double y,
 	}
 	pango_font_description_set_style (font_description, font_style);
 
-	pango_layout_set_text (pango_layout, string, -1);
+	pango_layout_set_text (pango_layout, string, n);
+
+	attrs = pango_attr_list_new ();
+#if 0
+	for (i = 0; i < n_dx; i++) {
+
+		attr = pango_attr_letter_spacing_new (pango_units_from_double (dx[i]));
+		attr->start_index = i;
+		if (i < n_dx - 1)
+			attr->end_index = i + 1;
+		else
+			attr->end_index = PANGO_ATTR_INDEX_TO_TEXT_END;
+
+		pango_attr_list_insert (attrs, attr);
+	}
+#endif
+	for (i = 0; i < n_dy; i++) {
+
+		attr = pango_attr_rise_new (-pango_units_from_double (dy[i]));
+		attr->start_index = i;
+		if (i < n_dy - 1)
+			attr->end_index = i + 1;
+		else
+			attr->end_index = PANGO_ATTR_INDEX_TO_TEXT_END;
+
+		pango_attr_list_insert (attrs, attr);
+	}
+	pango_layout_set_attributes (pango_layout, attrs);
+	pango_attr_list_unref (attrs);
+
 	pango_layout_set_font_description (pango_layout, font_description);
 	pango_layout_get_extents (pango_layout, &rectangle, NULL);
 
@@ -1072,7 +1204,7 @@ _lock_pango_layout (LsmSvgView *view)
 		view->pango_layout_stack = g_slist_prepend (view->pango_layout_stack, view->pango_layout);
 		view->pango_layout = pango_layout_new (pango_context);
 
-		lsm_debug_render ("[LsmSvgView::show_text] Create a new pango layout");
+		lsm_debug_render ("[LsmSvgView::_lock_pango_layout] Create a new pango layout");
 
 		return TRUE;
 	}
@@ -1087,7 +1219,7 @@ _unlock_pango_layout (LsmSvgView *view, gboolean need_pop)
 {
 
 	if (need_pop) {
-		lsm_debug_render ("[LsmSvgView::show_text] Free the child pango layout");
+		lsm_debug_render ("[LsmSvgView::_unlock_pango_layout] Free the child pango layout");
 
 		if (view->pango_layout != NULL) {
 			g_object_unref (view->pango_layout);
@@ -1096,23 +1228,31 @@ _unlock_pango_layout (LsmSvgView *view, gboolean need_pop)
 			view->pango_layout_stack = g_slist_delete_link (view->pango_layout_stack,
 									view->pango_layout_stack);
 		} else
-			g_warning ("[LsmSvgView::show_text] Pango layout stack empty");
+			g_warning ("[LsmSvgView::_unlock_pango_layout] Pango layout stack empty");
 	}
 
 	view->is_pango_layout_in_use = FALSE;
 }
 
-void
-lsm_svg_view_show_text (LsmSvgView *view, char const *string, double x, double y)
+static void
+_show_text (LsmSvgView *view,
+	    unsigned int n, char const *string, 
+	    unsigned int n_x, double *x, unsigned int n_y, double *y,
+	    unsigned int n_dx, double *dx, unsigned int n_dy, double *dy)
 {
 	LsmSvgViewPathInfos path_infos = default_path_infos;
+	PangoRectangle extents;
+	PangoLayoutIter *layout_iter;
 	const LsmSvgStyle *style;
 	gboolean need_pop;
+	double x_text;
+	double y_text;
+	double x_end, y_end;
+	double baseline;
+	cairo_t *cairo;
 
-	if (string == NULL || string[0] == '\0')
-		return;
 
-	g_return_if_fail (LSM_IS_SVG_VIEW (view));
+	cairo = view->dom_view.cairo;
 
 	style = view->style;
 
@@ -1120,28 +1260,143 @@ lsm_svg_view_show_text (LsmSvgView *view, char const *string, double x, double y
 
 	need_pop = _lock_pango_layout (view);
 
-	_update_pango_layout (view, string, x, y, &path_infos);
+	cairo_get_current_point (cairo, &x_text, &y_text);
+	if (x != NULL)
+		x_text = x[0];
+	if (y != NULL)
+		y_text = y[0];
+
+
+	_update_pango_layout (view, n, string, x_text, y_text, n_dx, dx, n_dy, dy, &path_infos);
 
 	if (style->writing_mode->value == LSM_SVG_WRITING_MODE_TB ||
 	    style->writing_mode->value == LSM_SVG_WRITING_MODE_TB_RL) {
 
-		cairo_save (view->dom_view.cairo);
-		cairo_rotate (view->dom_view.cairo, M_PI / 2.0);
-		cairo_move_to (view->dom_view.cairo, path_infos.extents.x1, path_infos.extents.y1);
+		cairo_save (cairo);
+		cairo_rotate (cairo, M_PI / 2.0);
+		cairo_move_to (cairo, path_infos.extents.x1, path_infos.extents.y1);
 
 		process_path (view, &path_infos);
 
-		cairo_restore (view->dom_view.cairo);
+		cairo_restore (cairo);
 	} else {
-		cairo_move_to (view->dom_view.cairo, path_infos.extents.x1, path_infos.extents.y1);
+		cairo_move_to (cairo, path_infos.extents.x1, path_infos.extents.y1);
 		process_path (view, &path_infos);
 	}
 
+	layout_iter = pango_layout_get_iter (view->pango_layout);
+	pango_layout_iter_get_line_extents (layout_iter, NULL, &extents);
+	baseline = pango_units_to_double (pango_layout_iter_get_baseline (layout_iter));
+	pango_layout_iter_free (layout_iter);
+
+	x_end = pango_units_to_double (extents.x + extents.width) + path_infos.extents.x1;
+	y_end = pango_units_to_double (extents.y) + baseline + path_infos.extents.y1;
+
+	if (view->debug_text) {
+		double x1, y1;
+
+		lsm_debug_render ("[LsmSvgView::_show_text] Logical extents %gx%g at %g,%g\n",
+			pango_units_to_double (extents.width),
+			pango_units_to_double (extents.height),
+			pango_units_to_double (extents.x),
+			pango_units_to_double (extents.y));
+
+		lsm_debug_render ("[LsmSvgView::_show_text] End point is %g, %g\n", x_end, y_end);
+		lsm_debug_render ("[LsmSvgView::_show_text] Baseline = %g\n", baseline);
+
+		x1 = 1;
+		y1 = 1;
+		cairo_device_to_user_distance (cairo, &x1, &y1);
+
+		cairo_set_source_rgba (cairo, 1.0, 0.0, 0.0, 0.2);
+		cairo_set_line_width (cairo, x1);
+		cairo_rectangle (cairo, path_infos.extents.x1, path_infos.extents.y1,
+				 pango_units_to_double (extents.x + extents.width),
+				 pango_units_to_double (extents.y + extents.height));
+		cairo_stroke (cairo);
+		cairo_set_source_rgba (cairo, 1.0, 0.0, 0.0, 0.2);
+		cairo_arc (cairo, x_end, y_end, 4.0 * x1, 0.0, 2.0 * M_PI);
+		cairo_arc (cairo, path_infos.extents.x1, path_infos.extents.y1, 4.0 * x1, 0.0, 2.0 * M_PI);
+		cairo_fill (cairo);
+	}
+
+	cairo_move_to (cairo, x_end, y_end);
+
 	_unlock_pango_layout (view, need_pop);
 }
 
 void
-lsm_svg_view_text_extents (LsmSvgView *view, char const *string, double x, double y, LsmExtents *extents)
+lsm_svg_view_start_text (LsmSvgView *view)
+{
+	g_return_if_fail (LSM_IS_SVG_VIEW (view));
+
+	cairo_move_to (view->dom_view.cairo, 0, 0);
+}
+
+void
+lsm_svg_view_end_text (LsmSvgView *view)
+{
+	g_return_if_fail (LSM_IS_SVG_VIEW (view));
+
+	cairo_new_path (view->dom_view.cairo);
+}
+
+void
+lsm_svg_view_show_text (LsmSvgView *view, char const *string, 
+			unsigned int n_x, double *x, unsigned int n_y, double *y,
+			unsigned int n_dx, double *dx, unsigned int n_dy, double *dy)
+{
+	unsigned int n, i;
+	char *iter = (char *) string;
+
+	if (string == NULL || string[0] == '\0')
+		return;
+
+	g_return_if_fail (LSM_IS_SVG_VIEW (view));
+	g_return_if_fail (n_x > 0 || x == NULL);
+	g_return_if_fail (n_y > 0 || y == NULL);
+	g_return_if_fail (n_dx > 0 || dx == NULL);
+	g_return_if_fail (n_dy > 0 || dy == NULL);
+
+	n = MAX (n_x, n_y);
+	if (n <= 1) {
+		_show_text (view, strlen (string), string, n_x, x, n_y, y, n_dx, dx, n_dy, dy);
+		return;
+	}
+
+	for (i = 0; (i < n - 1) && iter[0] != '\0'; i++) {
+		char *next_char;
+
+		next_char = g_utf8_next_char (iter);
+		_show_text (view, next_char - iter, iter,
+			    i < n_x ? n_x - i : 0,
+			    i < n_x ? &x[i] : NULL,
+			    i < n_y ? n_y - i : 0,
+			    i < n_y ? &y[i] : NULL,
+			    i < n_dx ? n_dx - i : 0,
+			    i < n_dx ? &dx[i] : NULL,
+			    i < n_dy ? n_dy - i : 0,
+			    i < n_dy ? &dy[i] : NULL);
+		iter = next_char;
+	}
+
+	if (iter[0] != '\0')
+		_show_text (view, strlen (iter), iter,
+			    i < n_x ? n_x - i : 0,
+			    i < n_x ? &x[i] : NULL,
+			    i < n_y ? n_y - i : 0,
+			    i < n_y ? &y[i] : NULL,
+			    i < n_dx ? n_dx - i : 0,
+			    i < n_dx ? &dx[i] : NULL,
+			    i < n_dy ? n_dy - i : 0,
+			    i < n_dy ? &dy[i] : NULL);
+}
+
+void
+lsm_svg_view_text_extents (LsmSvgView *view, char const *string,
+			   double x, double y,
+			   unsigned int n_dx, double *dx, unsigned int n_dy, double *dy,
+			   LsmExtents *extents)
 {
 	LsmSvgViewPathInfos path_infos = default_path_infos;
 	gboolean need_pop;
@@ -1159,7 +1414,7 @@ lsm_svg_view_text_extents (LsmSvgView *view, char const *string, double x, doubl
 
 	need_pop = _lock_pango_layout (view);
 
-	_update_pango_layout (view, string, x, y, &path_infos);
+	_update_pango_layout (view, strlen (string), string, x, y, n_dx, dx, n_dy, dy, &path_infos);
 
 	_unlock_pango_layout (view, need_pop);
 
@@ -2044,11 +2299,13 @@ lsm_svg_view_push_composition (LsmSvgView *view, LsmSvgStyle *style)
 	do_mask = (g_strcmp0 (style->mask->value, "none") != 0);
 	do_filter = (g_strcmp0 (style->filter->value, "none") != 0);
 
-	if ((view->style->opacity->value < 1.0 ||
-	     view->style->enable_background->value == LSM_SVG_ENABLE_BACKGROUND_NEW) &&
-	    !do_filter &&
-	    !view->is_clipping &&
-	    !view->style->ignore_group_opacity) {
+	if (G_UNLIKELY((view->style->opacity->value < 1.0 ||
+			view->style->enable_background->value == LSM_SVG_ENABLE_BACKGROUND_NEW ||
+			view->style->comp_op->value != LSM_SVG_COMP_OP_SRC_OVER) &&
+		       !do_filter &&
+		       !view->is_clipping &&
+		       !view->style->ignore_group_opacity &&
+		       view->dom_view.cairo != NULL)) {
 		LsmSvgViewBackground *background;
 
 		lsm_debug_render ("[LsmSvgView::push_composition] Push group");
@@ -2062,12 +2319,12 @@ lsm_svg_view_push_composition (LsmSvgView *view, LsmSvgStyle *style)
 		view->background_stack = g_list_prepend (view->background_stack, background);
 	}
 
-	if (do_clip) {
+	if (G_UNLIKELY (do_clip)) {
 		lsm_debug_render ("[LsmSvgView::push_style] Start clip '%s'", style->clip_path->value);
 		lsm_svg_view_push_clip (view);
 	}
 
-	if (do_mask) {
+	if (G_UNLIKELY (do_mask)) {
 		lsm_debug_render ("[LsmSvgView::push_style] Start mask '%s'", style->mask->value);
 		lsm_svg_view_push_mask (view);
 	}
@@ -2075,7 +2332,7 @@ lsm_svg_view_push_composition (LsmSvgView *view, LsmSvgStyle *style)
 	/* Don't do filtering during a clipping operation, as filter will
 	 * create a new subsurface, where clipping should occur with the path
 	 * of the clip-path element. */ 
-	if (do_filter && !view->is_clipping) {
+	if (G_UNLIKELY (do_filter && !view->is_clipping)) {
 		lsm_debug_render ("[LsmSvgView::push_style] Start filter '%s'", style->filter->value);
 		lsm_svg_view_push_filter (view);
 	}
@@ -2097,6 +2354,7 @@ void lsm_svg_view_pop_composition (LsmSvgView *view)
 	gboolean do_filter;
 	gboolean do_mask;
 	gboolean do_clip;
+	cairo_t *cairo;
 
 	g_return_if_fail (LSM_IS_SVG_VIEW (view));
 	g_return_if_fail (view->style != NULL);
@@ -2110,25 +2368,33 @@ void lsm_svg_view_pop_composition (LsmSvgView *view)
 	/* Don't do filtering during a clipping operation, as filter will
 	 * create a new subsurface, where clipping should occur with the path
 	 * of the clip-path element. */ 
-	if (do_filter && !view->is_clipping)
+	if (G_UNLIKELY (do_filter && !view->is_clipping))
 		lsm_svg_view_pop_filter (view);
 
-	if (do_mask)
+	if (G_UNLIKELY (do_mask))
 		lsm_svg_view_pop_mask (view);
 
-	if (do_clip)
+	if (G_UNLIKELY (do_clip))
 		lsm_svg_view_pop_clip (view);
 
-	if ((view->style->opacity->value < 1.0 ||
-	     view->style->enable_background->value == LSM_SVG_ENABLE_BACKGROUND_NEW) &&
-	    !do_filter &&
-	    !view->is_clipping &&
-	    !view->style->ignore_group_opacity) {
+	cairo = view->dom_view.cairo;
+
+	if (G_UNLIKELY ((view->style->opacity->value < 1.0 ||
+			 view->style->enable_background->value == LSM_SVG_ENABLE_BACKGROUND_NEW ||
+			 view->style->comp_op->value != LSM_SVG_COMP_OP_SRC_OVER) &&
+			!do_filter &&
+			!view->is_clipping &&
+			!view->style->ignore_group_opacity &&
+			cairo != NULL)) {
 		g_slice_free (LsmSvgViewBackground, view->background_stack->data);
 		view->background_stack = g_list_delete_link (view->background_stack, view->background_stack);
 
 		cairo_pop_group_to_source (view->dom_view.cairo);
-		cairo_paint_with_alpha (view->dom_view.cairo, view->style->opacity->value);
+		if (G_UNLIKELY (view->style->comp_op->value != LSM_SVG_COMP_OP_SRC_OVER))
+			lsm_cairo_set_comp_op (cairo, view->style->comp_op->value);
+		cairo_paint_with_alpha (cairo, view->style->opacity->value);
+		if (G_UNLIKELY (view->style->comp_op->value != LSM_SVG_COMP_OP_SRC_OVER))
+			lsm_cairo_set_comp_op (cairo, LSM_SVG_COMP_OP_SRC_OVER);
 		lsm_debug_render ("[LsmSvgView::pop_composition] Pop group");
 	}
 
@@ -2267,6 +2533,10 @@ lsm_svg_view_set_debug (LsmDomView *view, const char *feature, gboolean enable)
 		svg_view->debug_mask = enable;
 	else if (g_strcmp0 (feature, "pattern") == 0)
 		svg_view->debug_pattern = enable;
+	else if (g_strcmp0 (feature, "group") == 0)
+		svg_view->debug_group = enable;
+	else if (g_strcmp0 (feature, "text") == 0)
+		svg_view->debug_text = enable;
 }
 
 LsmSvgView *
